#' @name boxdesign
#' @rdname boxdesign
#' 
#' @title Create and modify a rectangular design
#'
#' @description
#' Rectangular ('box') design based on lower and upper
#' limits, with the option to distribute the points logarithmically
#' rather than linearly (for a positive lower limit).
#'
#' \code{extend.boxdesign} generates an additional design which is
#' space-filling with respect to the first.
#' \code{shrink.boxdesign} removes rows from a given design.
#'
#' There is an extraction method (\code{[}) for \code{boxdesign}
#' objects, and an \code{rbind} method.  Also \code{head} and
#' \code{tail} methods, which do not preserve the \code{boxdesign}
#' class (see Details).
#' 
#' @param n size of design; new points to add, points to retain, points to show
#' @param param named list of parameter lower and upper bounds
#' @param log. indices or names of parameters to have their points on a log scale
#' @param U underlying set of points on \eqn{[0, 1]^p}
#'
#' @return A design matrix with named rows and columns of class
#' \code{boxdesign}.  Has a \code{param} argument with the ranges,
#' and possibly a \code{log.} argument.
#'
#' @details In \code{boxdesign}, the argument \code{U} would
#' typically not be specified, as it is used mainly by
#' \code{extend.boxdesign}.  But it can be specified to override
#' the default arguments in \code{maximin}.
#'
#' \code{extend.boxdesign} is for adding rows.  The \code{update}
#' argument is for adjusting the ranges of parameters, and only needs
#' to be specified for parameters whose ranges change.
#' \code{shrink.boxdesign} is for removing rows.  An internal
#' function \code{tozeroone} is used to rescale \code{X} in both
#' cases.
#'
#' For a new object of class \code{boxdesign} comprising the first 10
#' rows, use \code{X[1:10, ]}.  But for a matrix of the first 10 rows,
#' eg as a quick summary, use \code{head(X, 10)}.
#'
#' @keywords utilities
#' @concept Experimental design
#' @seealso \code{\link{lhcube}} for space-filling designs on the unit cube.
NULL

#' @rdname boxdesign
#' @import utils
#' @export
#' @examples
#' X <- boxdesign(10, list(foo = c(1, 2), bar = c(3, 10)))
#' Y <- boxdesign(10, list(foo = c(1, 2), bar = c(3, 10)), log = "bar")
#'

boxdesign <- function(n, param = list(), log. = NULL, U = maximin(n, p)) {

    ## sort out the ranges

    nms <- names(param)
    if (is.null(nms))
        stop("Need names on \'param\'")
    p <- length(nms)
    U <- U # instantiate U

    range <- try(do.call("rbind", param), silent = TRUE)
    if (inherits(range, "try-error"))
        stop("Cannot interpret \'param\' as ranges")
    err <- range[, 1] >= range[, 2]
    if (any(err))
        stop("odd values in \'param\': ", paste(nms[err], collapse = ", "))

    ## sort out the logs

    if (!is.null(log.)) {
        if (is.numeric(log.)) {
            if (!all(log. %in% 1:p))
                stop("odd value for \'log.\'")
        } else {
            log. <- match(log., nms)
            if (any(is.na(log.)))
                stop("Unrecognised name(s) in \'log.\'")
        }

        range[log., ] <- log(range[log., ])
        if (any(is.na(range)))
            stop("Negative lower bound on log parameter(s)")
    }

    ## map back
    
    X <- sapply(1:p, function(j) qunif(U[, j], range[j, 1], range[j, 2]))
    if (!is.null(log.))
        X[, log.] <- exp(X[, log.])
    colnames(X) <- nms

    ## attach arguments as attributes and class
    
    attr(X, "param") <- param
    attr(X, "log.") <- log.
    class(X) <- "boxdesign"
    X
}

## simple mapper is internal

tozeroone <- function(X, param, log. = NULL) {
    
    X <- unclass(X)
    stopifnot(is.matrix(X), is.list(param),
              identical(colnames(X), names(param)),
              is.null(log.) || (is.numeric(log.) && log. %in% 1:ncol(X)))
    range <- do.call("rbind", param)
    if (!is.null(log.)) {
        X[, log.] <- log(X[, log.])
        range[log., ] <- log(range[log., ])
    }
    lower <- range[, 1]
    upper <- range[, 2]
    ones <- rep(1, nrow(X))
    U <- (X - (ones %o% lower)) / (ones %o% (upper - lower))
    U <- unclass(U)
    attr(U, "param") <- NULL
    attr(U, "log.") <- NULL
    U
}

#' @rdname boxdesign
#' @param X object of class \code{boxdesign}
#' @param update list of the parameter ranges to be updated
#' @param ... additional arguments to \code{extend.lhcube} or \code{shrink.lhcube}
#' @export
#' @examples
#' 
#' ## now extend and combine this design
#'
#' Z <- extend.boxdesign(Y, n = 5) # just 5 new rows
#'
#' ## change some limits
#' 
#' Z <- extend.boxdesign(Y, n = 5, update = list(bar = c(7, 11)))
#' Z <- rbind(Y, Z) # add them on to Y

extend.boxdesign <- function(X, n, update = list(), ...) {

    stopifnot(inherits(X, "boxdesign"))

    ## merge the parameter ranges: only the ranges to be updated are
    ## supplied in update

    param <- attr(X, "param")
    newparam <- jointparam <- param
    if (length(update) > 0) {
        nms <- names(update)
        if (is.null(nms) || !all(nms %in% names(param)))
            stop("Unmatched name in \'update\'")
        newparam[nms] <- update[nms]
        jointparam[nms] <- lapply(nms, function(nm) range(c(param[[nm]], update[[nm]])))
    }

    ## now generate points on [0, 1] under the new ranges
    
    log. <- attr(X, "log.")
    U <- tozeroone(X, newparam, log.)
    V <- extend.lhcube(U, n = n, ...)
    Y <- boxdesign(param = newparam, log. = log., U = V)
    rownames(Y) <- funames(n, avoid = rownames(X))
    attr(Y, "param") <- jointparam
    
    Y
}

    
## shrink box design

#' @rdname boxdesign
#' @export

shrink.boxdesign <- function(X, n, ...) {

    stopifnot(inherits(X, "boxdesign"))
    param <- attr(X, "param")
    log. <- attr(X, "log.")
    U <- tozeroone(X, param = param, log. = log.)
    ii <- shrink.lhcube(U, n, type = "indices", ...)
    X[ii, ]
}

## extraction method for boxdesign

#' @rdname boxdesign
#' @param x,y arguments to \code{[} operator or \code{rbind}
#' @export
#' @examples
#' ## use the extraction operator and rbind
#'
#' keep <- Z[, 1] < 1.5
#' Z[keep, ]
#' Z[keep, "bar"]
#' rbind(Z[1:5, ], Z[8:10, ])
#' 
#' ## example of shrinking a design
#'
#' Z <- boxdesign(100, param = list(foo = c(0, 1), bar = c(0, 1)))
#' keep <- Z[, "foo"] <= Z[, "bar"]
#' Z <- shrink.boxdesign(Z[keep, ], 20)
#'
#' ## head (and tail) just return matrices
#'
#' head(Z, 10)

"[.boxdesign" <- function(x, ...) {
    Y <- NextMethod("[", drop = FALSE)
    nms <- colnames(Y)
    attr(Y, "param") <- attr(x, "param")[nms]
    log. <- attr(x, "log.")
    if (!is.null(log.)) {
        mm <- intersect(colnames(x)[log.], nms)
        if (length(mm) > 0)
            attr(Y, "log.") <- match(mm, nms)
    }
    class(Y) <- "boxdesign"
    Y
}

#' @rdname boxdesign
#' @export

rbind.boxdesign <- function(x, y) {
    nms <- colnames(x)
    stopifnot(identical(nms, colnames(y)))
    Y <- rbind(unclass(x), unclass(y))

    ## and sort out the attributes

    parx <- attr(x, "param")
    pary <- attr(y, "param")
    param <- lapply(nms, function(nm) range(c(parx[[nm]], pary[[nm]])))
    names(param) <- nms
    attr(Y, "param") <- param
    
    log. <- union(attr(x, "log."), attr(y, "log."))
    if (!is.null(log.))
        attr(Y, "log.") <- sort(log.)
    class(Y) <- "boxdesign"
    Y
}

## these are usefully overloaded nb use unclass

#' @rdname boxdesign
#' @export

head.boxdesign <- function(x, n = 6L, ...) {
    head.matrix(unclass(x), n, ...)
}

#' @rdname boxdesign
#' @export

tail.boxdesign <- function(x, n = 6L, ...) {
    tail.matrix(unclass(x), n, ...)
}
