#' @name lhc
#' @rdname lhc
#' @title Create an n by p (maximin) Latin Hypercube design.
#'
#' @description
#' A Latin Hypercube (LHC) is a random space-filling design with
#' equally-spaced margins on the interval (0, 1).  It is often helpful
#' to perturb the points (\code{noisy = TRUE}) to introduce some
#' small-scale variability in the margins. Bad LHC designs are
#' possible; one approach to avoid this is to select a good design
#' from many candidates, such as a \code{maximin} LHC.
#'
#' Functions \code{extend.lhcube} and \code{shrink.lhcube} use the
#' maximin criterion to choose good extensions or subsets.  In both
#' cases, \code{n} indicates the number of rows in the new design.
#'
#' @param n number of rows in the design
#' @param p number of columns in the design
#' @param noisy logical, perturb points?
#'
#' @return A design matrix on \eqn{[0, 1]^p}.
#'
#' @section Warning:
#' Neither \code{extend.lhcube} nor
#' \code{shrink.lhcube} will provide a warning if \code{U} is not
#' contained in \eqn{[0, 1]^p}, as it is sometimes convenient to use a larger range (eg in \code{extend.boxdesign}).
#' 
#' @section Helpful tips:
#' Each row gets a unique name.  Pass this name along with the row
#' values into your code, as an additional input.  Make sure it gets
#' written into the output as well.  That way there can be no
#' confusion about which rows in the design are associated with which
#' outputs.
#'
#' To map the columns of the output to other distributions, use the
#' Probability Integral Transform; see the Examples.
#'
#' @keywords utilities
#' @concept Experimental design
#' @seealso The functions in the \code{DiceDesign} package.
#' \code{\link{boxdesign}} for creating simple rectangular designs.
NULL

#' @rdname lhc
#' @export
#' @examples
#' foo <- lhcube(10, 3)
#' print(foo)
#'
#' ## to map a column to a different range use the Probability
#' ## Integral Transform
#'
#' foo[, 2] <- qunif(foo[, 2], min = 10, max = 70)
#' foo[, 3] <- qnorm(foo[, 3], mean = 5.5, sd = 2.1)
#' print(foo)
#'

lhcube <- function(n, p, noisy = TRUE) {
    stopifnot(n > 0, p > 0, is.logical(noisy))
    robj <- sapply(1L:p, function(j) sample.int(n))
    dim(robj) <- c(n, p) # in case n = 1 or d = 1
    if (noisy) robj <- robj - runif(n * p)
    else robj <- robj - 0.5
    robj <- robj / n
    row.names(robj) <- funames(n)
    robj
}

#' @rdname lhc
#' @param ... Arguments to \code{lhcube}
#' @param nrep Number of of replications
#' @export
#' @examples
#'
#' ## maximin LHCs
#' 
#' lhc1 <- lhcube(21, 2, noisy = FALSE)
#' lhc2 <- maximin(21, 2, noisy = FALSE)
#' plot(lhc1, type = "p", pch = 20, col = 1)
#' points(lhc2, pch = 20, col = 2)
#' 

maximin <- function(..., nrep = 100) {
    best <- NULL
    ddbest <- 0
    for (i in 1L:nrep) {
        lhc <- lhcube(...)
        dd <- min(dist(lhc))
        if (dd > ddbest) {
            best <- lhc
            ddbest <- dd
        }
    }
    lhc
}

## add rows to a LHC

#' @rdname lhc
#' @param U Latin hypercube
#' @export
#' @examples
#'
#' ## extend LHCs using maximin
#'
#' lhc1 <- lhcube(21, 2)
#' lhc2 <- extend.lhcube(lhc1, n = 17)
#' plot(lhc1, type = "p", pch = 20, col = 1)
#' points(lhc2, pch = 20, col = 2)
#' 

## no warning for out of range, since this occurs in extend.boxdesign

extend.lhcube <- function(U, n, noisy = TRUE, nrep = 100) {
    U <- unclass(U)
    stopifnot(is.matrix(U), is.numeric(U))
#    if (any(U < 0 | U > 1))
#      warning("cols of \'U\' not on [0, 1], consider rescaling")
    p <- ncol(U)
    best <- NULL
    ddbest <- 0
    for (i in 1L:nrep) {
        lhc <- lhcube(n, p, noisy = noisy)
        dd <- min(dist(rbind(U, lhc)))
        if (dd > ddbest) {
            best <- lhc
            ddbest <- dd
        }
    }
    lhc
}

## delete rows from a LHC

#' @rdname lhc
#' @param type Type of object to return, either the selected rows of
#' \code{U} or the indices of the selected rows in \code{U}
#' @export
#' @examples
#'
#' ## drop rows from LHC using maximin
#'
#' lhc1 <- lhcube(21, 2)
#' lhc2 <- shrink.lhcube(lhc1, n = 10)
#' plot(lhc1, type = "p", pch = 20, col = 1)
#' points(lhc2, pch = 1, cex = 1.25, col = 2)

shrink.lhcube <- function(U, n, type = c("row", "indices"), nrep = 100) {
    U <- unclass(U)
    stopifnot(is.matrix(U), is.numeric(U))
    type <- match.arg(type)
#    if (any(U < 0 | U > 1))
#      warning("cols of \'U\' not on [0, 1], consider rescaling")
    m <- nrow(U); stopifnot(n < m)
    best <- numeric(0)
    ddbest <- 0
    for (i in 1L:nrep) {
        ii <- sample.int(m, n)
        lhc <- U[ii, , drop=FALSE]
        dd <- min(dist(lhc))
        if (dd > ddbest) {
            best <- ii
            ddbest <- dd
        }
    }
    if (type == "row")
        best <- U[best, , drop=FALSE]
    best
}
