#' Row echelon form
#'
#' Put m by n matrix \code{X} into row echelon form.
#'
#' @param X matrix
#' @param tol tolerance for testing equality with zero
#'
#' @return A matrix the same size as \code{X}, but with zeros below
#' the diagonal, and on the diagonal as well if \code{X} does not have
#' full column rank.
#'
#' @details See Strang (2006, sec 2.2) for details about row echelon
#' form.  \code{ref} implements a generalisation of the outer product
#' Gaussian elimination algorithm of Golub and Van Loan (1996,
#' Algorithm 3.2.1).
#'
#' @seealso This code is used in \code{\link{prvar}}.
#'
#' @references
#' G.H. Golub and C.F. Van Loan, 1996, Matrix Computations, 3rd ed,
#' The John Hopkins University Press, Baltimore, MD, USA.
#'
#' G. Strang, 2006, Linear Algebra and its Applications, 4th ed,
#' Thomson Brooks/Cole Belmont CA, USA.
#'
#' @export
#' @examples
#' ## test out a few obvious cases
#'
#' pivot <- function(X) {
#'   X <- X[!apply(X == 0, 1, all), , drop=FALSE]
#'   apply(X != 0, 1, which.max)
#' }
#' 
#' X <- matrix(rnorm(18), 6, 3)
#' Y <- ref(X)
#' print(Y); print(pivot(Y)) # == c(1, 2, 3)
#' ref(t(X))
#' print(Y); print(pivot(Y)) # == c(1, 2, 3)
#' X <- cbind(X[, 1:2], 3 * X[, 1] - 4 * X[, 2], X[, 3])
#' Y <- ref(X)
#' print(Y); print(pivot(Y)) # == c(1, 2, 4)
#' Y <- ref(X[1:3, ])
#' print(Y); print(pivot(Y)) # == c(1, 2, 4)

ref <- function(X, tol = 1e-8) {

    stopifnot(is.matrix(X), is.numeric(X), !is.na(X))
    m <- nrow(X); n <- ncol(X)

    i <- 1L
    for (j in 1L:n) {

        piv <- i - 1L + which.max(abs(X[i:m, j]))
        xpiv <- X[piv, j]
##        print(X)
##        print(c(i = i, j = j, piv = piv, xpiv = xpiv))
        if (abs(xpiv) < tol) {
            X[i:m, j] <- 0.0
            next
        }
        
        if (i < m) {
            rows <- (i + 1L):m
            if (j < n) {
                cols <- (j + 1L):n
                if (piv > i)
                    X[c(i, piv), ] <- X[c(piv, i), ]
                X[rows, j] <- X[rows, j] / xpiv
                X[rows, cols] <- X[rows, cols] - X[rows, j] %o% X[i, cols]
            }
            X[rows, j] <- 0.0
        }
        i <- i + 1L
        if (i > m) break
    }
    X
}
